Load PCFL_aHowe.

(*=CA *)
Definition CA : ExpRelations :=
  fun E t e f => exists R:ExpRelations,
  Compatible R /\ OPreAdequate R /\ R E t e f.

Lemma CA_refl : Reflexive CA.
Proof.
intros E t e.
exists IdRel; split.
  apply IdRel_comp.
  split. 
- apply IdRel_OPreAdeq.
 - reflexivity.
Qed.

Lemma CA_trans : Transitive CA.
Proof.
intros E t e f g H1 H2.
destruct H1; destruct H2.
exists (CompRel x x0); split.
  apply CompRel_comp; tauto.
  split. 
- apply CompRel_OPreAdeq; tauto.
 - exists f; tauto.
Qed.

Lemma CA_OPreAdeq : OPreAdequate CA.
Proof.
intros E t e f H.
destruct H as [R]; destruct H as [CR]; destruct H as [AR].
apply AR.
assumption.
Qed.

Lemma CA_comp : Compatible CA.
Proof.
repeat split; intros;
try (destruct H as [R H];
     destruct H as [CR H];
     destruct H as [AR H]);
try (destruct H0 as [R0 H0];
     destruct H0 as [CR0 H0];
     destruct H0 as [AR0 H0]);
try (destruct H1 as [R1 H1];
     destruct H1 as [CR1 H1];
     destruct H1 as [AR1 H1]);
try apply CA_refl;
eapply CA_trans;
  try (exists R;
       split; try assumption;
       split; try assumption;
       apply CR; try eassumption;
       eapply Comp_refl; assumption);
  try (exists R0;
       split; try assumption;
       split; try assumption;
       apply CR0; try eassumption;
       eapply Comp_refl; assumption);
eapply CA_trans;
  try (exists R0;
       split; try assumption;
       split; try assumption;
       apply CR0; try eassumption;
       eapply Comp_refl; assumption);
  try (exists R1;
       split; try assumption;
       split; try assumption;
       apply CR1; try eassumption;
       eapply Comp_refl; assumption).
Qed.

Definition GrCA : GrndRelations :=
  Exp_Grnd CA.
(*=End *)

(*=CA OaSim*)
Lemma OaSim_CA : OaSim <: CA.
Proof.
repeat intro.
exists OaSim.
split. 
- eapply OaSim_precongruence.
- split. 
* apply OaSim_OPreAdeq. 
* assumption.
Qed.

Lemma GrCA_aSim :  GrCA <: aSim.
Proof.
eapply IncldRel_trans; try apply PaSim_aSim.
pcofix GrCA_PaSim.
repeat intro; dependent destruction e1.
repeat destruct H0;
  destruct H; destruct H0.
rename x into R; rename e2 into t.
destruct t; pfold;
constructor; intros.
- (*case UNIT*)
  assert (Convrg e) by
    (eexists; eassumption).
  assert (Convrg f) by
    (rewrite <- ActionIdSub;
    eapply H0;
      try eassumption;
      rewrite ActionIdSub;
        assumption).
  destruct H4.
  assert (Val x) by (eapply Ev_Val_end; eassumption).
  dependent destruction H5. assumption.
- (*case LAM*)
(* show f converges to a LAM *)
  assert (Convrg e) by
    (eexists; eassumption).
  assert (Convrg f) by
    (rewrite <- ActionIdSub;
    eapply H0;
      try eassumption;
      rewrite ActionIdSub;
        assumption).
  destruct H4.
  assert (Val x) by (eapply Ev_Val_end; eassumption).
  dependent destruction H5.
  exists e0; split; try assumption; intro.
  right. apply GrCA_PaSim.
  eapply CA_trans.
  apply OaSim_CA;
    apply aSim_OaSim;
    apply Kle_aSim.
  repeat intro.
  eapply EvAPP.
    eapply H2.
    eassumption.
  eapply CA_trans.
  eapply CA_comp.
    exists R; eauto.
    eapply CA_refl.
  apply OaSim_CA;
    apply aSim_OaSim;
    apply Kle_aSim.
  repeat intro.
  dependent destruction H5.
  + assert (LAM e0 = LAM e1) by (eapply Determinacy; eassumption).
      dependent destruction H5.
      assumption.
  + inversion H5.
- (*case NIL*)
  assert (Convrg e) by
    (eexists; eassumption).
  assert (Convrg f) by
    (rewrite <- ActionIdSub;
    eapply H0;
      try eassumption;
      rewrite ActionIdSub;
        assumption).
  destruct H4.
  assert (Val x); try (eapply Ev_Val_end; eassumption).
  dependent destruction H5.
  + assumption.
  + assert (Conv:=ExConvrg [] t); assert (NConv:=ExNotConvrg [t; LIST t] t).
    destruct Conv as [conv Conv]; destruct NConv as [nconv NConv].
    specialize (Conv idSub); rewrite ActionIdSub in Conv.
    destruct Conv as [econv Conv].
    assert (R _ _ (CASE e conv nconv) (CASE f conv nconv)) by
      (repeat rewrite ActionIdSub;
      apply H;
      try apply Comp_refl;
      assumption).
    apply H0 in H5.
    assert (Convrg (STmExp idSub (CASE e conv nconv))) by
      (rewrite ActionIdSub;
      eexists;
      econstructor;
      eassumption).
    apply H5 in H6.
    rewrite ActionIdSub in H6.
    destruct H6; dependent destruction H6.
    * assert (H7:=Determinacy H4 H6_).
      inversion H7.
    * assert (Convrg (STmExp {|h;t|} nconv)) by
        (eexists; eassumption).
      apply NConv in H6.
      inversion H6.
    * inversion H6.
- (*case CONS*)
  assert (Convrg e) by
    (exists (CONS h t0); assumption).
  assert (Convrg f) by
    (rewrite <- ActionIdSub;
    eapply H0;
      try eassumption;
      rewrite ActionIdSub;
        assumption).
  destruct H4.
  assert (Val x); try (eapply Ev_Val_end; eassumption).
  dependent destruction H5.
  + assert (Conv:=ExConvrg [t; LIST t] t); assert (NConv:=ExNotConvrg [] t).
    destruct Conv as [conv Conv]; destruct NConv as [nconv NConv].
    specialize (Conv {|h; t0|}).
    destruct Conv as [econv Conv].
    assert (R _ _ (CASE e nconv conv) (CASE f nconv conv)) by
      (repeat rewrite ActionIdSub;
      apply H;
      try apply Comp_refl;
      assumption).
    apply H0 in H5.
    assert (Convrg (STmExp idSub (CASE e nconv conv))).
      (rewrite ActionIdSub;
      eexists;
      econstructor;
      try eassumption;
      eapply Conv).
    apply H5 in H6.
    rewrite ActionIdSub in H6.
    destruct H6; dependent destruction H6.
    * assert (Convrg (STmExp idSub nconv)) by
        (rewrite ActionIdSub; eexists; eassumption).
      apply NConv in H6.
      inversion H6.
    * assert (H7:=Determinacy H4 H6_).
      inversion H7.
    * inversion H6.
  + exists e0, e1.
    split; try assumption.
    split; right; apply GrCA_PaSim.
    * eapply CA_trans.
      apply OaSim_CA;
        apply aSim_OaSim;
        apply Kle_aSim.
      repeat intro.
      eapply (EvCASEC h (VAR (ZVAR _ _))).
        eapply H2.
        eassumption.
      eapply CA_trans.
      eapply CA_comp.
        exists R; eauto.
        eapply CA_refl.
        eapply CA_refl.
      apply OaSim_CA;
        apply aSim_OaSim;
        apply Kle_aSim.
      repeat intro.
      dependent destruction H5.
        assert (H6:=Determinacy H4 H5_).
          inversion H6.
        assert (CONS e0 e1 = CONS h0 t) by (eapply Determinacy; eassumption).
          dependent destruction H5.
          assumption.
        inversion H5.
    * eapply CA_trans.
      apply OaSim_CA;
        apply aSim_OaSim;
        apply Kle_aSim.
      repeat intro.
      eapply (EvCASEC t0 (VAR (SVAR _ (ZVAR _ _)))).
        eapply H2.
        eassumption.
      eapply CA_trans.
      eapply CA_comp.
        exists R; eauto.
        eapply CA_refl.
        eapply CA_refl.
      apply OaSim_CA;
        apply aSim_OaSim;
        apply Kle_aSim.
      repeat intro.
      dependent destruction H5.
        assert (H6:=Determinacy H4 H5_).
          inversion H6.
        assert (CONS e0 e1 = CONS h0 t3) by (eapply Determinacy; eassumption).
          dependent destruction H5.
          assumption.
        inversion H5.
Qed.

Lemma CA_OaSim : CA <: OaSim.
Proof.
intro E; induction E;
  intros t e f H.
- apply aSim_OaSim.
  apply GrCA_aSim.
  assumption.
- apply OaSim_LAMrev.
  apply IHE.
  apply CA_comp.
  assumption.
Qed.

Corollary CA_is_OaSim : CA <:> OaSim.
Proof.
unfold EqRel. split.
- apply CA_OaSim.
- apply OaSim_CA.
Qed.
(*=End *)