Load PCFL_gHowe.

(*=Ctx*)
Inductive Ctx E t : Env -> Ty -> Set :=
| CHOLE  : Ctx E t E t
| CCONS0 : forall E' t', Ctx E t E' t' -> Exp E' (LIST t') ->
                 Ctx E t E' (LIST t')
| CCONS1 : forall E' t', Exp E' t' -> Ctx E t E' (LIST t') ->
                 Ctx E t E' (LIST t')
| CCASE0 : forall E' t1 t2, Ctx E t E' (LIST t1) -> Exp E' t2 ->
                 Exp (t1 :: (LIST t1) :: E') t2 -> Ctx E t E' t2
| CCASE1 : forall E' t1 t2, Exp E' (LIST t1) -> Ctx E t E' t2 ->
                 Exp (t1 :: (LIST t1) :: E') t2 -> Ctx E t E' t2
| CCASE2 : forall E' t1 t2, Exp E' (LIST t1) -> Exp E' t2 ->
                 Ctx E t (t1 :: (LIST t1) :: E') t2 -> Ctx E t E' t2
| CLAM   : forall E' t1 t2, Ctx E t (t1 :: E') t2 -> Ctx E t E' (ARR t1 t2)
| CAPP0  : forall E' t1 t2, Ctx E t E' (ARR t1 t2) -> Exp E' t1 -> Ctx E t E' t2
| CAPP1  : forall E' t1 t2, Exp E' (ARR t1 t2) -> Ctx E t E' t1 -> Ctx E t E' t2
| CFIX   : forall E' t', Ctx E t (t' :: E') t' -> Ctx E t E' t'.

Fixpoint Rep {E t E' t'} (C:Ctx E t E' t') (e:Exp E t) : Exp E' t' :=
match C in Ctx _ _ E' t' return Exp E' t' with
| CHOLE _ _ => e
| CCONS0 C' f => CONS (Rep C' e) f
| CCONS1 f C' => CONS f (Rep C' e)
| CCASE0 C' f1 f2 => CASE (Rep C' e) f1 f2
| CCASE1 f0 C' f2 => CASE f0 (Rep C' e) f2
| CCASE2 f0 f1 C' => CASE f0 f1 (Rep C' e)
| CLAM C' => LAM (Rep C' e)
| CAPP0 C' f => APP (Rep C' e) f
| CAPP1 f C' => APP f (Rep C' e)
| CFIX C' => FIX (Rep C' e)
end.

Notation "C '@' e" := (Rep C e) (at level 41, right associativity).

Reserved Notation "C1 ;: C2" (at level 39).
Fixpoint CompCtx {E t E' t' E'' t''} (C':Ctx E' t' E'' t'') (C:Ctx E t E' t') :
  Ctx E t E'' t'' :=
match C' in Ctx _ _ E'' t'' return Ctx E t E'' t'' with
| CHOLE _ _ => C
| CCONS0 C_ f => CCONS0 (C_ ;: C) f
| CCONS1 f C_ => CCONS1 f (C_ ;: C)
| CCASE0 C_ f1 f2 => CCASE0 (C_ ;: C) f1 f2
| CCASE1 f0 C_ f2 => CCASE1 f0 (C_ ;: C) f2
| CCASE2 f0 f1 C_ => CCASE2 f0 f1 (C_ ;: C)
| CLAM C_ => CLAM (C_ ;: C)
| CAPP0 C_ f => CAPP0 (C_ ;: C) f
| CAPP1 f C_ => CAPP1 f (C_ ;: C)
| CFIX C_ => CFIX (C_ ;: C)
end
where "C1 ;: C2" := (CompCtx C1 C2).


Lemma CompCtx_Rep {E t E' t' E'' t''} :
  forall (C':Ctx E' t' E'' t'') (C:Ctx E t E' t') (e:Exp E t),
  C';:C @ e = C' @ C @ e.
Proof.
repeat intro; dependent induction C'; simpl;
  try rewrite IHC'; reflexivity.
Qed.

Lemma Sub_Ctx {E} : forall (s:GSub E) t, exists (C:Ctx E t _ t),
  forall e, @Kle tt _ (STmExp s e) (C @ e) /\
            @Kle tt _ (C @ e) (STmExp s e).
Proof.
induction E; repeat intro.
- exists (CHOLE _ _); simpl; intro.
  rewrite Grnd_GSub.
  split; apply Kle_refl.
- destruct (IHE (tlSub s) t) as [C H].
  exists (C ;: (CAPP0 (CLAM (CHOLE _ _)) (STmExp nilSub (hdSub s)))); intro.
  rewrite CompCtx_Rep.
  simpl.
  split; eapply Kle_trans; try apply H;
    simpl; eapply Kle_trans;
    try (apply Kle_APP; repeat econstructor);
  repeat rewrite <- ActionScS;
  erewrite ActionnilSub; rewrite ActionIdSub;
  rewrite <- consSub_ScS;
  rewrite <- Sub_consSub;
  apply Kle_refl.
Qed.

Lemma Comp_Ctx {E t E' t'} : forall R:ExpRelations, Compatible R ->
  forall (e f:Exp E t) (C:Ctx E t E' t'),
  R _ _ e f -> R _ _ (C@e) (C@f).
Proof.
repeat intro; rename H into CR; rename H0 into H.
assert (RR:Reflexive R) by (apply Comp_refl; assumption).
dependent induction C;
try (simpl; eapply CR; try eapply RR);
try apply IHC; assumption.
Qed.
(*=End*)

(*=ObOrd*)
Definition ObOrd : ExpRelations :=
  fun E t e f => forall C, @Kle () TOP (C@e) (C@f).

Lemma ObOrd_refl : Reflexive ObOrd.
Proof.
repeat intro.
assumption.
Qed.

Lemma ObOrd_trans : Transitive ObOrd.
Proof.
repeat intro.
apply H0.
apply H.
assumption.
Qed.

Lemma ObOrd_Ctx {E t E' t'} : forall (e f:Exp E t) (C:Ctx E t E' t'),
  ObOrd e f -> ObOrd (C@e) (C@f).
Proof.
repeat intro.
rewrite <- CompCtx_Rep;
  rewrite <- CompCtx_Rep in H0.
apply H.
assumption.
Qed.

Lemma ObOrd_comp : Compatible ObOrd.
Proof.
repeat intro; repeat split;
  intros; try apply ObOrd_refl.
- eapply ObOrd_trans.
  + assert (CONS e el = (CCONS0 (CHOLE _ _) el) @ e) by
      reflexivity.
    rewrite H1.
    eapply ObOrd_Ctx.
    eassumption.
  + simpl.
    assert (CONS f el = (CCONS1 f (CHOLE _ _)) @ el) by
      reflexivity.
    rewrite H1.
    assert (CONS f fl = (CCONS1 f (CHOLE _ _)) @ fl) by
      reflexivity.
    rewrite H2.
    eapply ObOrd_Ctx.
    assumption.
- eapply ObOrd_trans.
  eapply ObOrd_trans.
  + assert (CASE e en ec = (CCASE0 (CHOLE _ _) en ec) @ e) by
      reflexivity.
    rewrite H2.
    eapply ObOrd_Ctx.
    eassumption.
  + simpl.
    assert (CASE f en ec = (CCASE1 f (CHOLE _ _) ec) @ en) by
      reflexivity.
    rewrite H2.
    eapply ObOrd_Ctx.
    eassumption.
  + simpl.
    assert (CASE f fn ec = (CCASE2 f  fn (CHOLE _ _)) @ ec) by
      reflexivity.
    rewrite H2.
    assert (CASE f fn fc = (CCASE2 f  fn (CHOLE _ _)) @ fc) by
      reflexivity.
    rewrite H3.
    eapply ObOrd_Ctx.
    assumption.
- assert (LAM e = (CLAM (CHOLE _ _)) @ e) by
      reflexivity.
    rewrite H0.
    assert (LAM f = (CLAM (CHOLE _ _)) @ f) by
      reflexivity.
    rewrite H1.
    eapply ObOrd_Ctx.
    assumption.
- eapply ObOrd_trans.
  + assert (APP e ex = (CAPP0 (CHOLE _ _) ex) @ e) by
      reflexivity.
    rewrite H1.
    eapply ObOrd_Ctx.
    eassumption.
  + simpl.
    assert (APP f ex = (CAPP1 f (CHOLE _ _)) @ ex) by
      reflexivity.
    rewrite H1.
    assert (APP f fx = (CAPP1 f (CHOLE _ _)) @ fx) by
      reflexivity.
    rewrite H2.
    eapply ObOrd_Ctx.
    assumption.
- assert (FIX e = (CFIX (CHOLE _ _)) @ e) by
      reflexivity.
    rewrite H0.
    assert (FIX f = (CFIX (CHOLE _ _)) @ f) by
      reflexivity.
    rewrite H1.
    eapply ObOrd_Ctx.
    assumption.
Qed.
(*=End*)

(*=CgA*)
Definition gConvrgRel : GrndRelations :=
  fun u t e f => match t with
    | TOP => Kle e f
    | _ => True
    end.

Definition gPreAdequate (R:GrndRelations) :=
  R <: gConvrgRel.

Definition OgConvrgRel : ExpRelations :=
  Grnd_Exp gConvrgRel.

Definition OgPreAdequate (R:ExpRelations) :=
  R <: OgConvrgRel.

Lemma ObOrd_OgPreAdeq : OgPreAdequate ObOrd.
Proof.
intros E t e f H g.
destruct t; simpl; try apply I.
destruct (Sub_Ctx g TOP) as [C H0].
eapply Kle_trans; try apply H0.
eapply Kle_trans; try apply H;
  apply H0.
Qed.

Lemma CgA_ObOrd : forall R, Compatible R -> OgPreAdequate R -> R <: ObOrd.
Proof.
intros R CR AR E t e f H C.
fold (@gConvrgRel tt _ (C @ e) (C @ f)).
eapply Grnd_Exp_R; fold OgConvrgRel; unfold Exp_Grnd.
apply AR.
apply Comp_Ctx; assumption.
Qed.

Lemma ObOrd_CgA : Compatible ObOrd /\ OgPreAdequate ObOrd.
Proof.
split.
- apply ObOrd_comp.
- apply ObOrd_OgPreAdeq.
Qed.

(*=End*)

(*ObOrd OgSim*)
Lemma OgSim_ObOrd : OgSim <: ObOrd.
Proof.
repeat intro.
eapply (Comp_Ctx OgSim_comp _ _ C) in H.
apply OgSim_gSim in H.
dependent destruction H.
assert (Val g) by
  (eapply Ev_Val_end; eassumption);
  dependent destruction H1.
  apply H; assumption.
Qed.

Lemma ObOrd_gSim : Exp_Grnd ObOrd <: gSim.
Proof.
cofix ObOrd_gSim.
repeat intro; dependent destruction e1.
unfold Exp_Grnd in H.
destruct e2; econstructor.
- specialize (H (CHOLE _ _)); simpl in H.
  apply H.
- intro; apply ObOrd_gSim; unfold Exp_Grnd.
  intro.
  assert (forall e, APP e a = CAPP0 (CHOLE _ (ARR _ e2_2)) a @ e) by
    reflexivity; repeat rewrite H0; clear H0.
  repeat rewrite <- CompCtx_Rep.
  apply H.
- assert (Ncon:=ExNotConvrg [e2; LIST e2] TOP);
    destruct Ncon as [ncon Ncon].
  specialize (H (CCASE0 (CHOLE _ _) (UNIT _) ncon)); simpl in H.
  intro.
  assert (Ev (CASE e (UNIT []) ncon) (UNIT _)) by
    (repeat constructor; assumption).
  apply H in H1.
  dependent destruction H1.
  + assumption.
  + assert (Convrg (STmExp {|h; t|} ncon)) by
      (eexists; eassumption).
    apply Ncon in H1.
    inversion H1.
- assert (Ncon:=ExNotConvrg [] TOP);
    destruct Ncon as [ncon Ncon].
  assert (H0:=H (CCASE0 (CHOLE _ _) ncon (UNIT _))); simpl in H0.
  repeat intro.
  assert (Ev (CASE e ncon (UNIT _)) (UNIT _)) by
    (eapply EvCASEC; repeat constructor; eassumption).
  apply H0 in H2.
  dependent destruction H2.
  + assert (Convrg (STmExp idSub ncon)) by
      (rewrite ActionIdSub; eexists; eassumption).
    apply Ncon in H2.
    inversion H2.
  + repeat eexists.
    * eassumption.
    * apply ObOrd_gSim; unfold Exp_Grnd.
      eapply (ObOrd_Ctx (CCASE0 (CHOLE _ _) h (VAR (ZVAR _ e2)))) in H;
        simpl in H.
      eapply ObOrd_trans.
        eapply ObOrd_trans; try eassumption.
      eapply OgSim_ObOrd;
        eapply gSim_OgSim;
        eapply Kle_gSim.

      assert (h = STmExp {|h;t|} (VAR (ZVAR [LIST e2] e2))) by reflexivity;
        rewrite H2 at 1;
      eapply Kle_CASEC.
      assumption.

      (*repeat intro.
        eapply EvCASEC; try eassumption.*)

      eapply OgSim_ObOrd;
        eapply gSim_OgSim;
        eapply Kle_gSim.

      assert (h0 = STmExp {|h0;t0|} (VAR (ZVAR [LIST e2] e2))) by reflexivity;
        rewrite H2;
      eapply Kle_CASEC.
      assumption.

      (*repeat intro.
      dependent destruction H2.
        assert (D:=Determinacy).
          specialize (D _ _ _ H2_1 _ H2_).
          inversion D.
        assert (D:=Determinacy).
          specialize (D _ _ _ H2_1 _ H2_).
          dependent destruction D.
          eassumption.
        inversion H2.*)

    * apply ObOrd_gSim; unfold Exp_Grnd.
      eapply (ObOrd_Ctx (CCASE0 (CHOLE _ _) t (VAR (SVAR _ (ZVAR _ _))))) in H;
        simpl in H.
      eapply ObOrd_trans.
        eapply ObOrd_trans; try eassumption.
      eapply OgSim_ObOrd;
        eapply gSim_OgSim;
        eapply Kle_gSim.

      assert (t = STmExp {|h;t|} (VAR (SVAR e2 (ZVAR [] (LIST e2))))) by
        reflexivity;
        rewrite H2;
      eapply Kle_CASEC.
      assumption.

      (*repeat intro.
        eapply EvCASEC; try eassumption.*)

      eapply OgSim_ObOrd;
        eapply gSim_OgSim;
        eapply Kle_gSim.

      assert (t0 = STmExp {|h0;t0|} (VAR (SVAR e2 (ZVAR [] (LIST e2))))) by
        reflexivity;
        rewrite H2;
      eapply Kle_CASEC.
      assumption.

      (*repeat intro.
      dependent destruction H2.
        assert (D:=Determinacy).
          specialize (D _ _ _ H2_1 _ H2_).
          inversion D.
        assert (D:=Determinacy).
          specialize (D _ _ _ H2_1 _ H2_).
          dependent destruction D.
          eassumption.
        inversion H2.*)

Qed.

Lemma ObOrd_OgSim : ObOrd <: OgSim.
Proof.
intro E; induction E;
  intros t e f H.
- apply gSim_OgSim.
  apply ObOrd_gSim.
  assumption.
- apply OgSim_LAMrev.
  apply IHE.
  apply ObOrd_comp.
  assumption.
Qed.
Corollary ObOrd_is_OgSim : ObOrd <:> OgSim.
Proof.
unfold EqRel. split.
- apply ObOrd_OgSim .
- apply OgSim_ObOrd .
Qed.
(*=End*)