Load PCFL_Relprop.

(*=Howe*)
Inductive Howe R `{CuS_Preorder R} : forall E t, Exp E t -> Exp E t -> Prop :=
| HVAR  : forall E t (v:Var E t) e, R _ _ (VAR v) e -> Howe (VAR v) e
| HUNIT : forall E (e:Exp E TOP), R _ _ (UNIT _) e -> Howe (UNIT _) e
| HNIL  : forall E t (e:Exp E (LIST t)), R _ _ (NIL _ _) e -> Howe (NIL _ _) e
| HCONS : forall E t (e f:Exp E t) el fl g, Howe e f -> Howe el fl -> R _ _ (CONS f fl) g ->
                     Howe (CONS e el) g
| HCASE : forall E t1 t2 (e f:Exp E (LIST t1)) (en fn:Exp E t2) ec fc g,
                     Howe e f -> Howe en fn -> Howe ec fc -> R _ _ (CASE f fn fc) g ->
                     Howe (CASE e en ec) g
| HLAM  : forall E t1 t2 (e f:Exp (t1::E) t2) g, Howe e f -> R _ _ (LAM f) g -> Howe (LAM e) g
| HAPP  : forall E t1 t2 (e f:Exp E (ARR t1 t2)) ex fx g,
                     Howe e f -> Howe ex fx -> R _ _ (APP f fx) g -> Howe (APP e ex) g
| HFIX  : forall E t (e f:Exp (t::E) t) g, Howe e f -> R _ _ (FIX f) g -> Howe (FIX e) g.

Lemma Howe_semitrans : forall R `{CuS_Preorder R} E t (e:Exp E t) f g,
                       Howe e f -> R E t f g -> Howe e g.
Proof.
intros.
dependent destruction H0; econstructor;
try eassumption; try eapply trans; eassumption.
Qed.

Lemma Howe_refl R `{CuS_Preorder R} : Reflexive (@Howe R _).
Proof.
unfold Reflexive; dependent induction e;
  econstructor; try apply refl; assumption.
Qed.

Lemma Howe_comp R `{CuS_Preorder R} : Compatible (@Howe R _).
Proof.
unfold Compatible; repeat split; intros; econstructor;
  try apply refl; assumption.
Qed.

Lemma R_Howe R `{CuS_Preorder R} : IncldRel R (@Howe R _).
Proof.
repeat intro; dependent destruction e; econstructor;
try eassumption; apply Howe_refl.
Qed.

Lemma Howe_Ren R `{CP:CuS_Preorder R} : CuR (@Howe R CP).
Proof.
assert (cur : CuR R); try (apply CuS_CuR; apply cus).
intros E t e f E' r H; generalize E' r; clear E' r;
  induction H; repeat intro; simpl; econstructor.
- assert (VAR (r t v) = RTmExp r (VAR v)); try reflexivity; rewrite H0; clear H0.
  apply cur. apply H.
- assert (UNIT E' = RTmExp r (UNIT E)); try reflexivity; rewrite H0; clear H0.
  apply cur. apply H.
- assert (NIL E' t = RTmExp r (NIL E t)); try reflexivity; rewrite H0; clear H0.
  apply cur. apply H.
- apply IHHowe1.
- apply IHHowe2.
- assert (CONS (RTmExp r f) (RTmExp r fl) = RTmExp r (CONS f fl));
    try reflexivity; rewrite H2; clear H2.
  apply cur. apply H1.
- apply IHHowe1.
- apply IHHowe2.
- apply IHHowe3.
- assert (CASE (RTmExp r f) (RTmExp r fn) (RTmExp (RTmL (RTmL r)) fc) =
    RTmExp r (CASE f fn fc));
    try reflexivity; rewrite H3; clear H3.
  apply cur. apply H2.
- apply IHHowe.
- assert (LAM (RTmExp (RTmL r) f) = RTmExp r (LAM f)); try reflexivity; rewrite H1; clear H1.
  apply cur. apply H0.
- apply IHHowe1.
- apply IHHowe2.
- assert (APP (RTmExp r f) (RTmExp r fx) = RTmExp r (APP f fx));
    try reflexivity; rewrite H2; clear H2.
  apply cur. apply H1.
- apply IHHowe.
- assert (FIX (RTmExp (RTmL r) f) = RTmExp r (FIX f)); try reflexivity; rewrite H1; clear H1.
  apply cur. apply H0.
Qed.

Lemma Howe_Lift R `{CP:CuS_Preorder R} : forall E t (e f:Exp E t), @Howe R _ _ _ e f ->
                    forall t0, Howe (ShTmExp t0 e) (ShTmExp t0 f).
Proof.
intros.
apply Howe_Ren.
apply H.
Qed.

Lemma Howe_STmL R `{CP:CuS_Preorder R} : forall E E' (s:Sub E E') s', SubRel (@Howe R _) s s' ->
                      forall t, SubRel (@Howe R _) (@STmL E E' t s) (STmL s').
Proof.
unfold SubRel; intros.
dependent destruction v.
- simpl; apply Howe_refl.
- simpl; apply Howe_Lift; apply H.
Qed.

Lemma Howe_Sub R `{CP:CuS_Preorder R} : Substitutive (@Howe R _).
Proof.
unfold Substitutive; repeat intro; dependent induction H; intros;
  eapply Howe_semitrans; try (eapply cus; eassumption);
  simpl; try eapply Howe_comp;
  try eapply IHHowe; try eapply IHHowe1;
    try eapply IHHowe2; try eapply IHHowe3;
  repeat apply Howe_STmL;
  try apply H0; assumption.
Qed.
(*=End*)