Load PCFL_syntax.

(*=Basic Rel Props*)
Definition Relations E1 E2 (A:E1 -> E2 -> Type) := forall e1 e2, A e1 e2 -> A e1 e2 -> Prop.

Definition IdRel {E1 E2} {A:E1 -> E2 -> Type} : Relations A :=
  fun e1 e2 e f => e = f.

Definition IncldRel {E1 E2} {A:E1 -> E2 -> Type} (R R':Relations A) :=
  forall e1 e2 e f, R e1 e2 e f -> R' _ _ e f.
Notation "R1 '<:' R2"  := (IncldRel R1 R2) (at level 50, left associativity).

Definition EqRel {E1 E2} {A:E1 -> E2 -> Type} (R R':Relations A) :=
R <: R' /\ R' <: R.
Notation "R1 '<:>' R2"  := (EqRel R1 R2) (at level 50, left associativity).

Lemma IncldRel_trans {E1 E2} {A:E1 -> E2 -> Type} : forall R R' R'' : Relations A,
       R <: R' -> R' <: R'' -> R <: R''.
Proof.
repeat intro.
apply H0.
apply H.
assumption.
Qed.

Definition CompRel {E1 E2} {A:E1 -> E2 -> Type} (R R':Relations A) : Relations A :=
  fun e1 e2 e f => exists g, R e1 e2 e g /\ R' _ _ g f.
Notation "R1 ';;' R2"  := (CompRel R1 R2) (at level 45, left associativity).

Definition InvRel {E1 E2} {A:E1 -> E2 -> Type} (R:Relations A) : Relations A :=
  fun e1 e2 e f => R e1 e2 f e.

Definition Reflexive {E1 E2} {A:E1 -> E2 -> Type} (R:Relations A) :=
  forall e1 e2 e, R e1 e2 e e.

Definition Refl {E1 E2} {A:E1 -> E2 -> Type} (R:Relations A) := 
   IdRel <: R.

Lemma Rdef {E1 E2} {A:E1 -> E2 -> Type} (R:Relations A) : 
forall e1 e2 e, Refl R -> R e1 e2 e e.
Proof.
unfold Refl; unfold IncldRel; unfold IdRel.
intros.
specialize (H e1 e2 e e). apply H. reflexivity.
Qed.

Lemma R_refl_IdRel {E1 E2} {A:E1 -> E2 -> Type} (R:Relations A) :
  Reflexive R <-> IdRel <: R.
Proof.
unfold Reflexive; unfold IncldRel; unfold IdRel.
split; intros.
- rewrite H0; apply H.
- apply H; reflexivity.
Qed.

Definition Symmetric {E1 E2} {A:E1 -> E2 -> Type} (R:Relations A) :=
  forall e1 e2 e f, R e1 e2 e f -> R e1 e2 f e.

Lemma R_sym_InvRel {E1 E2} {A:E1 -> E2 -> Type} (R:Relations A) :
  Symmetric R <-> InvRel R <: R.
Proof.
unfold Symmetric; unfold IncldRel; unfold InvRel.
split; intros.
- apply H; assumption.
- apply H; assumption.
Qed.

Definition Transitive {E1 E2} {A:E1 -> E2 -> Type} (R:Relations A) :=
  forall e1 e2 e f g, R e1 e2 e f -> R _ _ f g -> R _ _ e g.

Lemma R_trans_CompRel {E1 E2} {A:E1 -> E2 -> Type} (R:Relations A) :
  Transitive R <-> R ;; R <: R.
Proof.
unfold Transitive; unfold IncldRel; unfold CompRel.
split; intros.
- destruct H0. eapply H; eapply H0.
- apply H; exists f. split; assumption.
Qed.

Definition Preorder {E1 E2} {A:E1 -> E2 -> Type} (R:Relations A) :=
  Reflexive R /\ Transitive R.
(*=End*)

(*=ExpRelations*)
Definition ExpRelations := Relations Exp.

Definition CuR (R:ExpRelations) :=
  forall E t e f E' r, R E t e f -> R E' t (RTmExp r e) (RTmExp r f).

Definition CuS (R:ExpRelations) :=
  forall E t e f E' s, R E t e f -> R E' t (STmExp s e) (STmExp s f).

Lemma CuS_CuR : forall R, CuS R -> CuR R.
Proof.
unfold CuS; unfold CuR; intros.
repeat rewrite Action_Ren_Sub.
apply H; apply H0.
Qed.

Class CuS_Preorder (R:ExpRelations) :=
  {refl  : Reflexive R;
   trans : Transitive R;
   cus   : CuS R}.
(*=End *)

(*=Compatible*)
Definition Compatible (R:ExpRelations) := forall E,
  (forall t v, R E t (VAR v) (VAR v)) /\
  (R E _ (UNIT _) (UNIT _)) /\
  (forall t, R E (LIST t) (NIL _ _) (NIL _ _)) /\
  (forall t e f el fl, R E t e f -> R _ _ el fl -> R _ _ (CONS e el) (CONS f fl)) /\
  (forall t1 t2 e f en fn ec fc, R E (LIST t1) e f -> R _ t2 en fn ->
                     R _ _ ec fc -> R _ _ (CASE e en ec) (CASE f fn fc)) /\
  (forall t1 t2 e f, R (t1::E) t2 e f -> R _ _ (LAM e) (LAM f)) /\
  (forall t1 t2 e f ex fx, R E (ARR t1 t2) e f -> R _ _ ex fx -> R _ _ (APP e ex) (APP f fx)) /\
  (forall t e f, R (t::E) t e f -> R _ _ (FIX e) (FIX f)).

Lemma IdRel_comp : Compatible IdRel.
Proof.
repeat split; intros;
  try rewrite H; try rewrite H0; try rewrite H1; try rewrite H2;
  reflexivity.
Qed.

Lemma Comp_refl {R:ExpRelations} : Compatible R -> Reflexive R.
Proof.
intros H E t e.
dependent induction e;
  eapply H; assumption.
Qed.

Lemma CompRel_comp : forall R R':ExpRelations, Compatible R -> Compatible R' ->
             Compatible ( R ;; R').
Proof.
repeat split; intros; unfold CompRel.
- exists (VAR v); split.
  eapply H. eapply H0.
- exists (UNIT E); split.
  eapply H. eapply H0.
- exists (NIL _ _); split.
  eapply H. eapply H0.
- destruct H1; destruct H2.
  exists (CONS x x0); split.
  eapply H; tauto.
  eapply H0; tauto.
- destruct H1; destruct H2; destruct H3.
  exists (CASE x x0 x1); split.
  eapply H; tauto.
  eapply H0; tauto.
- destruct H1.
  exists (LAM x); split.
  eapply H; tauto.
  eapply H0; tauto.
- destruct H1; destruct H2.
  exists (APP x x0); split.
  eapply H; tauto.
  eapply H0; tauto.
- destruct H1.
  exists (FIX x); split.
  eapply H; tauto.
  eapply H0; tauto.
Qed.

Definition Precongruence (R:ExpRelations) :=
  Transitive R /\ Compatible R.
(*=End*)

(*=GrndRelations*)
Definition GrndRelations := Relations (fun (_:()) => Exp []).

Definition Exp_Grnd : ExpRelations -> GrndRelations :=
  fun R a => R [].

Definition Grnd_Exp : GrndRelations -> ExpRelations :=
  fun R E t e f => forall (g:GSub E), R tt _ (STmExp g e) (STmExp g f).
Hint Unfold Grnd_Exp.

(* unused *)
Lemma Exp_Grnd_IncldRel : forall R R',  R <: R' ->
                         (Exp_Grnd R) <: (Exp_Grnd R').
Proof.
repeat intro.
apply H.
apply H0.
Qed.
(* unused *)
Lemma Grnd_Exp_IncldRel : forall R R',  R <: R' ->
                         (Grnd_Exp R) <: (Grnd_Exp R').
Proof.
repeat intro.
auto.
Qed.

(* unused *)
Lemma Exp_Grnd_refl {R:ExpRelations} : Reflexive R -> Reflexive (Exp_Grnd R).
Proof.
unfold Reflexive.  intros. apply H.
Qed.

Lemma Grnd_Exp_refl {R:GrndRelations} : Reflexive R -> Reflexive (Grnd_Exp R).
Proof.
unfold Reflexive. unfold Grnd_Exp. intros.
apply H.
Qed.

(* unused *)
Lemma Exp_Grnd_trans {R:ExpRelations} : Transitive R -> Transitive (Exp_Grnd R).
Proof.
unfold Transitive; unfold Exp_Grnd; intros.
eapply H.
  apply H0.
  apply H1.
Qed.

(* see: PCFL_O*Sim.v: *)
Lemma Grnd_Exp_trans {R:GrndRelations} : Transitive R -> Transitive (Grnd_Exp R).
Proof.
unfold Transitive; unfold Grnd_Exp; intros.
eapply H.
  apply H0.
  apply H1.
Qed.
(* see: PCFL_O*Sim.v: *)
Lemma Grnd_Exp_Subclosed {R:GrndRelations} : CuS (Grnd_Exp R).
Proof.
repeat intro.
repeat rewrite <- ActionScS.
apply H.
Qed.
(* see: PCFL_O*Sim.v: *)

Lemma R_Grnd_Exp {R:GrndRelations} :  R <: (Exp_Grnd (Grnd_Exp R)).
Proof.
repeat intro; destruct e1.
repeat rewrite Grnd_GSub.
assumption.
Qed.

(* see: PCFL_O*Sim.v: *)
Lemma Grnd_Exp_R {R:GrndRelations} :  (Exp_Grnd (Grnd_Exp R)) <: R.
Proof.
repeat intro; destruct e1.
specialize (H idSub).
repeat rewrite ActionIdSub in H.
assumption.
Qed.
(* unused *)
Lemma  R_Grnd_Exp_Eq {R:GrndRelations} :  R <:> (Exp_Grnd (Grnd_Exp R)).
Proof.
unfold EqRel. split.
- apply R_Grnd_Exp.
- apply Grnd_Exp_R.
Qed.


(* unused *)
Lemma R_Exp_Grnd {R:ExpRelations}: CuS R -> R  <:  (Grnd_Exp (Exp_Grnd R)).
Proof.
repeat intro; unfold Exp_Grnd.
apply H. assumption.
Qed.


(*=End*)

(*=Kleene*)
Definition Kle : GrndRelations :=
  fun u t e f => forall g, Ev e g -> Ev f g.

Lemma Kle_refl : Reflexive Kle.
Proof.
repeat intro; assumption.
Qed.

Lemma Kle_trans : Transitive Kle.
Proof.
repeat intro.
apply H0.
apply H.
assumption.
Qed.

Lemma Kle_APP : forall t1 t2 (e:Exp _ _) (f:Exp _ _) (p:Exp [t1] t2),
  Ev e (LAM p) ->
  (@Kle tt _ (APP e f) (STmExp {|f|} p) /\
  @Kle tt _ (STmExp {|f|} p) (APP e f)).
Proof.
split; repeat intro.
- dependent destruction H0;
    try dependent destruction H0.
  assert (H0:=Determinacy H H0_);
    dependent destruction H0.
  assumption.
- repeat econstructor;
  eassumption.
Qed.

Lemma Kle_CASEN : forall t1 t2 (e:Exp _ _) (en:Exp _ t2) (ec:Exp _ _),
  Ev e (NIL _ t1) ->
  (@Kle tt _ (CASE e en ec) en /\
  @Kle tt _ en (CASE e en ec)).
Proof.
split; repeat intro.
- dependent destruction H0;
    try dependent destruction H0.
  assumption.
  assert (H0:=Determinacy H H0_);
    dependent destruction H0.
- repeat econstructor;
  assumption.
Qed.

Lemma Kle_CASEC : forall t1 t2 (e:Exp _ _) (en:Exp _ t2) (ec:Exp _ _)
                               (ex:Exp _ t1) (el:Exp _ _),
  (Ev e (CONS ex el) ->
  (@Kle tt _ (CASE e en ec) (STmExp {|ex;el|} ec) /\
  @Kle tt _ (STmExp {|ex;el|} ec) (CASE e en ec))).
Proof.
split; repeat intro.
- dependent destruction H0;
    try dependent destruction H0.
  assert (H0:=Determinacy H H0_);
    dependent destruction H0.
  assert (H0:=Determinacy H H0_);
    dependent destruction H0.
  assumption.
- eapply EvCASEC;
  repeat econstructor;
  eassumption.
Qed.

Lemma Kle_FIX : forall t (e:Exp _ t),
  @Kle tt _ (FIX e) (STmExp {|FIX e|} e) /\
  @Kle tt _ (STmExp {|FIX e|} e) (FIX e).
Proof.
split; repeat intro.
- dependent destruction H.
  assumption.
  dependent destruction H.
- constructor.
  assumption.
Qed.


(* unused*)
(*
Lemma Ev_Kle : forall t (e f:Exp [] t), Ev e f -> @Kle tt t e f.
Proof.
unfold Kle; intros.
erewrite Determinacy.
- apply EvVal; eapply Ev_Val_end; eassumption.
- eassumption.
- assumption.
Qed.
*)
(*=End*)

(*=PreAdequacy*)
(*
Definition CR  t (e f : Exp [] t) :=
  Convrg e -> Convrg f. *)
Definition ConvrgRel : GrndRelations :=
  fun u t e f => Convrg e -> Convrg f.

Lemma ConvrgRel_refl : Reflexive ConvrgRel.
Proof.
repeat intro; assumption.
Qed.

Lemma ConvrgRel_trans : Transitive ConvrgRel.
Proof.
repeat intro; apply H0; apply H; assumption.
Qed.

Definition PreAdequate (R:GrndRelations):=
   R <: ConvrgRel.

Hint Unfold Convrg.

Lemma Kle_preadeq : PreAdequate Kle.
Proof.
intro; intros.
unfold ConvrgRel. 
intro; destruct H0.
exists x; apply H; assumption.
Qed.

Definition OConvrgRel : ExpRelations :=
  Grnd_Exp ConvrgRel.

Definition OPreAdequate R:=
  R <: OConvrgRel.

(* PCFL_CA.v:*)
Lemma IdRel_OPreAdeq : OPreAdequate IdRel.
Proof.
repeat intro.
rewrite <- H.
assumption.
Qed.

Lemma CompRel_OPreAdeq : forall R R':ExpRelations,
             OPreAdequate R -> OPreAdequate R' ->
             OPreAdequate (R ;; R').
Proof.
repeat intro.
destruct H1.
eapply H0. apply H1.
eapply H. apply H1.
assumption.
Qed.
(* unused *)
Lemma OConvrgRel_refl : Reflexive OConvrgRel.
Proof.
intros E t e g H.
assumption.
Qed.

Lemma OConvrgRel_trans : Transitive OConvrgRel.
Proof.
intros E t e f g H0 H1 s H.
apply H1.
apply H0.
assumption.
Qed.
(*=End *)

(*=Substitutivity*)
Definition RenRelations := Relations Ren.

Definition SubRelations := Relations Sub.

Definition RenRel : ExpRelations -> RenRelations:=
fun R E E' r r' => forall t v, R E' t (VAR (r t v)) (VAR (r' t v)).

Definition SubRel : ExpRelations -> SubRelations:=
fun R E E' s s' => forall t v, R E' t (s t v) (s' t v).

(* unused *)
Lemma RenRel_refl {R:ExpRelations} : Reflexive R -> Reflexive (RenRel R).
Proof.
unfold Reflexive; unfold RenRel; intros; apply H.
Qed.
(* see PCFL_*Howe.*)
Lemma SubRel_refl {R:ExpRelations} : Reflexive R -> Reflexive (SubRel R).
Proof.
unfold Reflexive; unfold SubRel; intros; apply H.
Qed.

Lemma RenRel_trans {R:ExpRelations} : Transitive R -> Transitive (RenRel R).
Proof.
unfold Transitive; unfold RenRel; intros; eapply H; try apply H0; apply H1.
Qed.

Lemma SubRel_trans {R:ExpRelations} : Transitive R -> Transitive (SubRel R).
Proof.
unfold Transitive; unfold SubRel; intros; eapply H; try apply H0; apply H1.
Qed.

Definition Substitutive (R:ExpRelations) :=
  forall E t e f, R E t e f -> forall E' s s',
                      (SubRel R) E E' s s' -> R _ _ (STmExp s e) (STmExp s' f).
(*=End *)