Load PCFL_OaSim.

(*=Howe aSim*)
Definition aHowe := @Howe OaSim OaSim_CuSPreord.

Lemma aHowe_mim_aSim :
  (forall e:Exp [] TOP, aHowe (UNIT _) e -> Ev e (UNIT _)) /\
  (forall t (e:Exp [] (LIST t)), (aHowe (NIL _ _) e -> Ev e (NIL _ _)) /\
    forall f fl, (aHowe (CONS f fl) e ->
    exists g gl, (Ev e (CONS g gl) /\ aHowe f g /\ aHowe fl gl))) /\
  (forall s t (e:Exp [] (ARR s t)) f, aHowe (LAM f) e ->
    exists g, (Ev e (LAM g) /\
    forall x, aHowe (STmExp {|x|} f) (STmExp {|x|} g))).
(*By inversion on the Howe relation and deﬁnition of similarity,
  using semi-transitivity and, in the lambda-case,
  substitutivity of the Howe relation.*)
Proof.
repeat split; intros.
- dependent destruction H. apply OaSim_aSim in H.
  dependent destruction H. apply H. repeat constructor.
- dependent destruction H. apply OaSim_aSim in H.
  dependent destruction H. apply H. repeat constructor.
- dependent destruction H. apply OaSim_aSim in H1.
  dependent destruction H1.
  assert (Ev (CONS f0 fl0) (CONS f0 fl0)); repeat constructor.
  apply H2 in H3; repeat destruct H3; destruct H4.
  exists x, x0; repeat split;
    try assumption;
    eapply Howe_semitrans; try apply aSim_OaSim; eassumption.
- dependent destruction H. apply OaSim_aSim in H0.
  dependent destruction H0.
  assert (Ev (LAM f0) (LAM f0)); repeat constructor.
  apply H0 in H1; destruct H1;
    exists x; split;
    try apply H1; intro;
    eapply Howe_semitrans.
  apply Howe_Sub; try apply H;
    unfold SubRel; intros.
  apply Howe_refl.
  apply aSim_OaSim.
  apply H1.
Qed.

Lemma aHowe_downclosed : forall t (e f g:Exp [] t),
  aHowe e f -> Ev e g -> aHowe g f.
Proof.
intros.
dependent induction H0; try assumption; dependent destruction H.
- eapply IHEv2.
  eapply Howe_semitrans; try eassumption.
  apply aSim_OaSim; apply OaSim_aSim in H2; dependent destruction H2;
  constructor; intros.
  + apply H2.
    constructor.
    apply aHowe_mim_aSim.
    eapply IHEv1; assumption.
    assumption.
  + apply H2.
    constructor.
    apply aHowe_mim_aSim.
    eapply IHEv1; assumption.
    assumption.
  + apply H3.
    constructor.
    apply aHowe_mim_aSim.
    eapply IHEv1; assumption.
    assumption.
  + apply H2.
    constructor.
    apply aHowe_mim_aSim.
    eapply IHEv1; assumption.
    assumption.
- apply IHEv2.
  assert (exists g gl, (Ev f0 (CONS g gl) /\ Howe h g /\ Howe t gl));
    try (apply aHowe_mim_aSim;
    eapply IHEv1; try reflexivity; try assumption);
    repeat destruct H3; destruct H4.
  assert (Howe (STmExp {|h; t|} e2) (STmExp {|x0; x1|} fc));
    try (apply Howe_Sub; try assumption; unfold SubRel; intros;
    repeat (dependent destruction v; simpl); assumption).
  eapply Howe_semitrans; try eassumption.
  apply aSim_OaSim; apply OaSim_aSim in H2; dependent destruction H2;
  constructor; intros.
  + apply H2.
    econstructor; eassumption.
  + apply H2.
    econstructor; eassumption.
  + apply H3.
    econstructor; eassumption.
  + apply H2.
    econstructor; eassumption.
- apply IHEv2.
  assert (H2 := IHEv1 _ H); dependent destruction H2; apply OaSim_aSim in H3;
    dependent destruction H3;
    assert (Ev (LAM f1) (LAM f1)); repeat constructor; apply H3 in H4;
    repeat destruct H4.
  assert (Howe (STmExp {|e2|} e) (STmExp {|fx|} f1));
    try (apply Howe_Sub; try assumption; unfold SubRel; intros;
    repeat (dependent destruction v0; simpl); try assumption).
  eapply Howe_semitrans; try eassumption.
  apply aSim_OaSim; eapply aSim_trans; try apply (H5 fx).
  apply OaSim_aSim in H1; dependent destruction H1;
  constructor; intros.
  + apply H1.
    econstructor; eassumption.
  + apply H1.
    econstructor; eassumption.
  + apply H2.
    econstructor; eassumption.
  + apply H1.
    econstructor; eassumption.
- apply IHEv.
  assert (Howe (STmExp {|FIX e|} e) (STmExp {|FIX f0|} f0));
    try (apply Howe_Sub; try assumption; unfold SubRel; intros;
    repeat (dependent destruction v; simpl);
    econstructor; try eassumption; apply OaSim_refl).
  eapply Howe_semitrans; try eassumption.
  apply aSim_OaSim;
  apply OaSim_aSim in H0; dependent destruction H0;
  constructor; intros.
  + apply H0.
    econstructor; eassumption.
  + apply H0.
    econstructor; eassumption.
  + apply H1.
    econstructor; eassumption.
  + apply H0.
    econstructor; eassumption.
Qed.

Lemma aHowe_aSim : (Exp_Grnd aHowe) <: aSim.
Proof.
cofix aHowe_aSim.
intros u t; intros.
destruct t; constructor; intros.
- apply aHowe_mim_aSim.
  eapply aHowe_downclosed; eassumption.
- assert (Howe (LAM e') f);
  try (eapply aHowe_downclosed; eassumption);
  apply aHowe_mim_aSim in H1.
  destruct H1.
  exists x; split.
  apply H1.
  intro.
  apply aHowe_aSim.
  apply H1.
- apply aHowe_mim_aSim.
  eapply aHowe_downclosed; eassumption.
- assert (Howe (CONS h t0) f);
  try (eapply aHowe_downclosed; eassumption);
  apply aHowe_mim_aSim in H1.
  repeat destruct H1.
  exists x, x0; repeat split;
  try apply aHowe_aSim; try apply H1; apply H2.
Qed.

Theorem aHowe_OaSim : aHowe <: OaSim.
Proof.
repeat intro.
apply aHowe_aSim;
  unfold Exp_Grnd.
apply Howe_Sub.
(* - assumption.
- apply SubRel_refl. *)
  apply Howe_refl.
Qed.

Corollary aHowe_is_OaSim : aHowe <:> OaSim.
Proof.
unfold EqRel. split.
- apply aHowe_OaSim .
- apply R_Howe.
Qed.

Theorem OaSim_precongruence : Precongruence OaSim.
Proof.
split.
- apply trans.
- unfold Compatible. repeat split;
  intros; apply aHowe_OaSim; apply Howe_comp;
  apply R_Howe; assumption.
Qed.
(*=End*)