Load PCFL_CA.

(*=Ctx*)
Inductive Ctx E t : Env -> Ty -> Set :=
| CHOLE  : Ctx E t E t
| CCONS0 : forall E' t', Ctx E t E' t' -> Exp E' (LIST t') ->
                 Ctx E t E' (LIST t')
| CCONS1 : forall E' t', Exp E' t' -> Ctx E t E' (LIST t') ->
                 Ctx E t E' (LIST t')
| CCASE0 : forall E' t1 t2, Ctx E t E' (LIST t1) -> Exp E' t2 ->
                 Exp (t1 :: (LIST t1) :: E') t2 -> Ctx E t E' t2
| CCASE1 : forall E' t1 t2, Exp E' (LIST t1) -> Ctx E t E' t2 ->
                 Exp (t1 :: (LIST t1) :: E') t2 -> Ctx E t E' t2
| CCASE2 : forall E' t1 t2, Exp E' (LIST t1) -> Exp E' t2 ->
                 Ctx E t (t1 :: (LIST t1) :: E') t2 -> Ctx E t E' t2
| CLAM   : forall E' t1 t2, Ctx E t (t1 :: E') t2 -> Ctx E t E' (ARR t1 t2)
| CAPP0  : forall E' t1 t2, Ctx E t E' (ARR t1 t2) -> Exp E' t1 -> Ctx E t E' t2
| CAPP1  : forall E' t1 t2, Exp E' (ARR t1 t2) -> Ctx E t E' t1 -> Ctx E t E' t2
| CFIX   : forall E' t', Ctx E t (t' :: E') t' -> Ctx E t E' t'.

Fixpoint Rep {E t E' t'} (C:Ctx E t E' t') (e:Exp E t) : Exp E' t' :=
match C in Ctx _ _ E' t' return Exp E' t' with
| CHOLE _ _ => e
| CCONS0 C' f => CONS (Rep C' e) f
| CCONS1 f C' => CONS f (Rep C' e)
| CCASE0 C' f1 f2 => CASE (Rep C' e) f1 f2
| CCASE1 f0 C' f2 => CASE f0 (Rep C' e) f2
| CCASE2 f0 f1 C' => CASE f0 f1 (Rep C' e)
| CLAM C' => LAM (Rep C' e)
| CAPP0 C' f => APP (Rep C' e) f
| CAPP1 f C' => APP f (Rep C' e)
| CFIX C' => FIX (Rep C' e)
end.

Notation "C '@' e" := (Rep C e) (at level 41, right associativity).

Reserved Notation "C1 ;: C2" (at level 39).
Fixpoint CompCtx {E t E' t' E'' t''} (C':Ctx E' t' E'' t'') (C:Ctx E t E' t') :
  Ctx E t E'' t'' :=
match C' in Ctx _ _ E'' t'' return Ctx E t E'' t'' with
| CHOLE _ _ => C
| CCONS0 C_ f => CCONS0 (C_ ;: C) f
| CCONS1 f C_ => CCONS1 f (C_ ;: C)
| CCASE0 C_ f1 f2 => CCASE0 (C_ ;: C) f1 f2
| CCASE1 f0 C_ f2 => CCASE1 f0 (C_ ;: C) f2
| CCASE2 f0 f1 C_ => CCASE2 f0 f1 (C_ ;: C)
| CLAM C_ => CLAM (C_ ;: C)
| CAPP0 C_ f => CAPP0 (C_ ;: C) f
| CAPP1 f C_ => CAPP1 f (C_ ;: C)
| CFIX C_ => CFIX (C_ ;: C)
end
where "C1 ;: C2" := (CompCtx C1 C2).


Lemma CompCtx_Rep {E t E' t' E'' t''} :
  forall (C':Ctx E' t' E'' t'') (C:Ctx E t E' t') (e:Exp E t),
  C';:C @ e = C' @ C @ e.
Proof.
repeat intro; dependent induction C'; simpl;
  try rewrite IHC'; reflexivity.
Qed.

Lemma Sub_Ctx_Kle {E} : forall (s:GSub E) t, exists (C:Ctx E t _ t),
  forall e, @Kle tt _ (STmExp s e) (C @ e) /\
            @Kle tt _ (C @ e) (STmExp s e).
Proof.
induction E; repeat intro.
- exists (CHOLE _ _); simpl; intro.
  rewrite Grnd_GSub.
  split; apply Kle_refl.
- destruct (IHE (tlSub s) t) as [C H].
  exists (C ;: (CAPP0 (CLAM (CHOLE _ _)) (STmExp nilSub (hdSub s)))); intro.
  rewrite CompCtx_Rep.
  simpl.
  split; eapply Kle_trans; try apply H;
    simpl; eapply Kle_trans;
    try (apply Kle_APP; repeat econstructor);
  repeat rewrite <- ActionScS;
  erewrite ActionnilSub; rewrite ActionIdSub;
  rewrite <- consSub_ScS;
  rewrite <- Sub_consSub;
  apply Kle_refl.
Qed.


Lemma Sub_Ctx_ConvrgRel {E} : forall (s:GSub E) t, exists (C:Ctx E t _ t),
  forall e, @ConvrgRel tt _ (STmExp s e) (C @ e) /\
            @ConvrgRel tt _ (C @ e) (STmExp s e).
Proof.
repeat intro.
destruct (Sub_Ctx_Kle s t).
exists x; split;
  apply Kle_preadeq;
  apply H.
Qed.

Lemma Comp_Ctx {E t E' t'} : forall R:ExpRelations, Compatible R ->
  forall (e f:Exp E t) (C:Ctx E t E' t'),
  R _ _ e f -> R _ _ (C@e) (C@f).
Proof.
repeat intro; rename H into CR; rename H0 into H.
assert (RR:Reflexive R) by (apply Comp_refl; assumption).
dependent induction C;
try (simpl; eapply CR; try eapply RR);
try apply IHC; assumption.
Qed.
(*=End*)

(*=aObOrd*)
Definition aObOrd : ExpRelations :=
  fun E t e f => forall t' C, @ConvrgRel () t' (C@e) (C@f).

Lemma aObOrd_refl : Reflexive aObOrd.
Proof.
repeat intro.
assumption.
Qed.

Lemma aObOrd_trans : Transitive aObOrd.
Proof.
repeat intro.
apply H0.
apply H.
assumption.
Qed.

Lemma aObOrd_Ctx {E t E' t'} : forall (e f:Exp E t) (C:Ctx E t E' t'),
  aObOrd e f -> aObOrd (C@e) (C@f).
Proof.
repeat intro.
rewrite <- CompCtx_Rep;
  rewrite <- CompCtx_Rep in H0.
apply H.
assumption.
Qed.

Lemma aObOrd_comp : Compatible aObOrd.
Proof.
repeat intro; repeat split;
  intros; try apply aObOrd_refl.
- eapply aObOrd_trans.
  + assert (CONS e el = (CCONS0 (CHOLE _ _) el) @ e) by
      reflexivity.
    rewrite H1.
    eapply aObOrd_Ctx.
    eassumption.
  + simpl.
    assert (CONS f el = (CCONS1 f (CHOLE _ _)) @ el) by
      reflexivity.
    rewrite H1.
    assert (CONS f fl = (CCONS1 f (CHOLE _ _)) @ fl) by
      reflexivity.
    rewrite H2.
    eapply aObOrd_Ctx.
    assumption.
- eapply aObOrd_trans.
  eapply aObOrd_trans.
  + assert (CASE e en ec = (CCASE0 (CHOLE _ _) en ec) @ e) by
      reflexivity.
    rewrite H2.
    eapply aObOrd_Ctx.
    eassumption.
  + simpl.
    assert (CASE f en ec = (CCASE1 f (CHOLE _ _) ec) @ en) by
      reflexivity.
    rewrite H2.
    eapply aObOrd_Ctx.
    eassumption.
  + simpl.
    assert (CASE f fn ec = (CCASE2 f  fn (CHOLE _ _)) @ ec) by
      reflexivity.
    rewrite H2.
    assert (CASE f fn fc = (CCASE2 f  fn (CHOLE _ _)) @ fc) by
      reflexivity.
    rewrite H3.
    eapply aObOrd_Ctx.
    assumption.
- assert (LAM e = (CLAM (CHOLE _ _)) @ e) by
      reflexivity.
    rewrite H0.
    assert (LAM f = (CLAM (CHOLE _ _)) @ f) by
      reflexivity.
    rewrite H1.
    eapply aObOrd_Ctx.
    assumption.
- eapply aObOrd_trans.
  + assert (APP e ex = (CAPP0 (CHOLE _ _) ex) @ e) by
      reflexivity.
    rewrite H1.
    eapply aObOrd_Ctx.
    eassumption.
  + simpl.
    assert (APP f ex = (CAPP1 f (CHOLE _ _)) @ ex) by
      reflexivity.
    rewrite H1.
    assert (APP f fx = (CAPP1 f (CHOLE _ _)) @ fx) by
      reflexivity.
    rewrite H2.
    eapply aObOrd_Ctx.
    assumption.
- assert (FIX e = (CFIX (CHOLE _ _)) @ e) by
      reflexivity.
    rewrite H0.
    assert (FIX f = (CFIX (CHOLE _ _)) @ f) by
      reflexivity.
    rewrite H1.
    eapply aObOrd_Ctx.
    assumption.
Qed.
(*=End*)

(*=CA*)
Lemma aObOrd_OPreAdeq : OPreAdequate aObOrd.
Proof.
intros E t e f H g.
destruct (Sub_Ctx_ConvrgRel g t) as [C H0].
eapply ConvrgRel_trans; try apply H0.
eapply ConvrgRel_trans; try apply H;
  apply H0.
Qed.

Lemma CA_aObOrd : CA <: aObOrd.
Proof.
intros E t e f H t' C.
destruct H as [R H]; destruct H as [CR H]; destruct H as [AR H].
eapply Grnd_Exp_R; fold OConvrgRel; unfold Exp_Grnd.
apply AR.
apply Comp_Ctx; assumption.
Qed.

Lemma aObOrd_CA : aObOrd <: CA.
Proof.
repeat intro.
eexists; split;
  try apply aObOrd_comp;
  split;
  try apply aObOrd_OPreAdeq.
assumption.
Qed.

Theorem aObOrd_is_CA : aObOrd <:> CA.
Proof.
split.
- apply aObOrd_CA.
- apply CA_aObOrd.
Qed.
(*=End*)