Load PCFL_Howe.

(*=aSim*)
CoInductive aSim : GrndRelations :=
| aSimTOP  : forall u (e f:Exp _ TOP),
             (Ev e (UNIT _) -> Ev f (UNIT _)) -> @aSim u _ e f
| aSimLIST : forall u t (e f:Exp _ (LIST t)),
             (Ev e (NIL _ _) -> Ev f (NIL _ _)) ->
             (forall h t, Ev e (CONS h t) ->
             (exists h' t', Ev f (CONS h' t') /\
                            @aSim u _ h h' /\
                            @aSim u _ t t')) ->
             @aSim u _ e f
| aSimARR  : forall u s t (e f:Exp _ (ARR s t)),
             (forall e', Ev e (LAM e') ->
             (exists f', Ev f (LAM f') /\
                         forall g,
                         @aSim u _ (STmExp {|g|} e') (STmExp {|g|} f'))) ->
             @aSim u _ e f.

(* Hint Constructors aSim. *)


Lemma aSim_refl : Reflexive aSim.
Proof.
cofix aSimr.
unfold Reflexive; intro.
intros t e.
 destruct t; econstructor; eauto.
Qed.

Lemma aSim_trans : Transitive aSim.
Proof.
cofix aSimt.
unfold Transitive.
intros.
dependent destruction H.
- dependent destruction H0; apply aSimTOP; intro; auto.
- dependent destruction H1; apply aSimLIST.
  + auto.
  + intros; apply H0 in H3. destruct H3; destruct H3; destruct H3. apply H2 in H3.
    destruct H3; destruct H3; destruct H3.
    exists x1, x2; split.
    * apply H3.
    * destruct H5; destruct H4. split; eapply aSimt; eauto.
- dependent destruction H0; apply aSimARR; intros.
  apply H in H1. destruct H1; destruct H1.
  apply H0 in H1. destruct H1; destruct H1.
  exists x0; split.
  + apply H1.
  + intro; eapply aSimt; eauto.
Qed.

Lemma Kle_aSim : IncldRel Kle aSim.
Proof.
intros u t.
induction t; intros; constructor; intros.
- apply H; assumption.
- exists e'; split; intros.
  + apply H; apply H0.
  + apply aSim_refl.
- apply H; apply H0.
- exists h, t0; repeat split;
  try apply aSim_refl.
  apply H; apply H0.
Qed.

Lemma Ev_aSim : forall t (e f:Exp [] t), Ev e f -> @aSim tt _ e f.
Proof.
intros.
apply Kle_aSim; repeat intro.
erewrite Determinacy.
- apply EvVal; eapply Ev_Val_end; eassumption.
- eassumption.
- assumption.
Qed.

Lemma aSim_PreAdeq : PreAdequate aSim.
Proof.
repeat intro; destruct H0.
assert (Val x) by (eapply Ev_Val_end; eassumption).
dependent destruction H1; dependent destruction H.
- eexists; eapply H; assumption.
- eexists; eapply H; assumption.
- eapply H0 in H1; repeat destruct H1.
  eexists; eassumption.
- eapply H in H0; repeat destruct H0.
  eexists; eassumption.
Qed.
(*=End *)

(*=Paco aSim*)
Require Import Paco.paco.
Inductive RaSim (R:GrndRelations) : GrndRelations :=
| RaSimTOP  : forall u (e f:Exp _ TOP),
              (Ev e (UNIT _) -> Ev f (UNIT _)) -> @RaSim R u _ e f
| RaSimLIST : forall u t (e f:Exp _ (LIST t)),
              (Ev e (NIL _ _) -> Ev f (NIL _ _)) ->
              (forall h t, Ev e (CONS h t) ->
              (exists h' t', Ev f (CONS h' t') /\
                             R u _ h h' /\
                             R u _ t t')) ->
              @RaSim R u _ e f
| RaSimARR  : forall u s t (e f:Exp _ (ARR s t)),
              (forall e', Ev e (LAM e') ->
              (exists f', Ev f (LAM f') /\
                          forall g, R u _ (STmExp {|g|} e') (STmExp {|g|} f'))) ->
              @RaSim R u _ e f.

Definition PaSim u t e f := paco4 RaSim bot4 u t e f.
Hint Unfold PaSim.
Lemma RaSim_mon: monotone4 RaSim.
Proof.
repeat intro.
destruct x0; destruct x1;
constructor; dependent destruction IN.
- assumption.
- repeat intro.
  specialize (H e' H0).
  repeat destruct H.
    exists x; split; try assumption.
  intro.
  apply LE.
  apply H1.
- assumption.
- repeat intro.
  specialize (H0 h t H1).
  repeat destruct H0.
    exists x, x0; split; try assumption.
  split;
    apply LE;
    apply H2.
Qed.
Hint Resolve RaSim_mon : paco.
Lemma PaSim_fold : PaSim <: RaSim PaSim.
Proof.
repeat intro.
destruct e1; induction e2;
constructor;
punfold H; dependent destruction H.
- assumption.
- intros.
  specialize (H _ H0);
    repeat destruct H.
  exists x. split.
  + assumption.
  + intro.
    unfold upaco4 in H1; unfold PaSim.
    specialize (H1 g); destruct H1.
    * assumption.
    * inversion H1.
- assumption.
- intros.
  specialize (H0 _ _ H1);
    repeat destruct H0.
  exists x, x0. split; try split.
  + assumption.
  + unfold upaco4 in H2; unfold PaSim.
    destruct H2; destruct H2.
    * assumption.
    * inversion H2.
  + unfold upaco4 in H2; unfold PaSim.
    destruct H2; destruct H3.
    * assumption.
    * inversion H3.
Qed.

Lemma aSim_PaSim : aSim <: PaSim.
Proof.
pcofix aSim_PaSim.
repeat intro; destruct e1.
pfold.
destruct e2; dependent destruction H0; constructor.
- assumption.
- repeat intro.
  specialize (H e' H0).
  repeat destruct H.
    exists x; split; try assumption.
  intro.
  right.
  apply aSim_PaSim.
  apply H1.
- assumption.
- repeat intro.
  specialize (H0 h t H1).
  repeat destruct H0.
    exists x, x0; split; try assumption.
  split;
    right;
    apply aSim_PaSim;
    apply H2.
Qed.

Lemma PaSim_aSim : PaSim <: aSim.
Proof.
cofix PaSim_aSim.
repeat intro.
destruct e1; destruct e2; constructor;
  punfold H; dependent destruction H.
- assumption.
- intros.
  specialize (H _ H0);
    repeat destruct H.
  exists x. split.
  + assumption.
  + intro; apply PaSim_aSim.
    unfold upaco4 in H1; unfold PaSim.
    specialize (H1 g); destruct H1.
    * assumption.
    * inversion H1.
- assumption.
- intros.
  specialize (H0 _ _ H1);
    repeat destruct H0.
  exists x, x0. split; try split.
  + assumption.
  + apply PaSim_aSim.
    unfold upaco4 in H2; unfold PaSim.
    destruct H2; destruct H2.
    * assumption.
    * inversion H2.
  + apply PaSim_aSim.
    unfold upaco4 in H2; unfold PaSim.
    destruct H2; destruct H3.
    * assumption.
    * inversion H3.
Qed.
Corollary aSim_is_PaSim: PaSim <:> aSim.
unfold EqRel. split.
- apply PaSim_aSim.
- apply aSim_PaSim.
Qed.
(*=End*)